@@ -52,6 +52,8 @@
 #include "file.h"
 #include "socket.h"
 #include "worker.h"
+#include "logger.h"
+#include "appender_std.h"
 #include "conf.h"
 #include "util.h"
 #include "module.h"
@@ -84,7 +86,7 @@ typedef struct self_s {
 #define SELF_FLAGS_NONE 0
 #define SELF_FLAGS_SKIP_COOKIE_FIRST_TIME 1
 #define SELF_FLAGS_GOT_SET_COOKIE 2
-  int log_mode;
+  logger_t *logger;
 } self_t;
 
 typedef struct request_s {
@@ -119,6 +121,7 @@ apr_status_t tcp_module_init(global_t *global);
  ***********************************************************************/
 char *none = "";
 int new_session = 0;
+int success = 1;
 
 /**
  * Helper
@@ -269,7 +272,7 @@ static apr_status_t wait_request(worker_t * worker, request_t *r) {
   peeklen = worker->socket->peeklen;
   worker->socket->peeklen = 0;
   if ((status = sockreader_new(&sockreader, worker->socket->transport,
-                               worker->socket->peek, peeklen, r->pool)) != APR_SUCCESS) {
+                               worker->socket->peek, peeklen)) != APR_SUCCESS) {
     goto out_err;
   }
 
@@ -374,7 +377,7 @@ static apr_status_t wait_response(worker_t * worker, response_t *r) {
   peeklen = worker->socket->peeklen;
   worker->socket->peeklen = 0;
   if ((status = sockreader_new(&sockreader, worker->socket->transport,
-                               worker->socket->peek, peeklen, r->pool)) != APR_SUCCESS) {
+                               worker->socket->peek, peeklen)) != APR_SUCCESS) {
     goto out_err;
   }
 
@@ -652,7 +655,7 @@ static apr_status_t do_body(self_t *self, worker_t *worker, apr_pool_t *p,
       }
 
       if (chunked) { 
-	line = apr_psprintf(worker->pbody, "%x\r\n", len);
+	line = apr_psprintf(worker->pbody, "%x\r\n", (unsigned int)len);
 	if ((status = worker_socket_send(worker, line, strlen(line))) != APR_SUCCESS) {
 	  return status;
 	}
@@ -806,21 +809,22 @@ static void *APR_THREAD_FUNC proxy_thread(apr_thread_t * thread, void *selfv) {
   worker_t *client = self->client;
   
   memset(&global, 0, sizeof(global));
-  global.log_mode = self->log_mode;
+  global.logger = self->logger;
   global.socktmo = 1000 * 300000;
   global.modules = apr_hash_make(self->pool);
   global.blocks = apr_hash_make(self->pool);
   
-  if ((status = worker_new(&server, "", "", &global, NULL)) != APR_SUCCESS) {
-    apr_thread_exit(thread, status);
-  } 
+  worker_new(&server, "", &global, NULL);
+
 
   if ((status = call_command(client, command_TIMEOUT, "_TIMEOUT", "300000")) != APR_SUCCESS) {
-    goto unlock;
+    apr_thread_mutex_unlock(self->mutex);
+    return NULL;
   }
 
   if ((status = call_command(server, command_TIMEOUT, "_TIMEMOUT", "300000")) != APR_SUCCESS) {
-    goto unlock;
+    apr_thread_mutex_unlock(self->mutex);
+    return NULL;
   }
 
   /* as long we are connected */
@@ -830,7 +834,6 @@ static void *APR_THREAD_FUNC proxy_thread(apr_thread_t * thread, void *selfv) {
     request.pool = pool;
     apr_pool_create(&pool, NULL);
     response.pool = pool;
-
     apr_pool_create(&ptmp, NULL);
 
     /* wait client request */
@@ -1161,7 +1164,7 @@ int proxy(self_t *self) {
 
   memset(&global, 0, sizeof(global));
   global.pool = self->pool;
-  global.log_mode = self->log_mode;
+  global.logger = self->logger;
   global.socktmo = 1000 * 300000;
   global.modules = apr_hash_make(self->pool);
   global.blocks = apr_hash_make(self->pool);
@@ -1172,9 +1175,7 @@ int proxy(self_t *self) {
 
   tcp_module_init(&global);
 
-  if ((status = worker_new(&listener, "", "", &global, NULL)) != APR_SUCCESS) {
-    return status;
-  } 
+  worker_new(&listener, "", &global, NULL);
 
   listener->listener_port = self->port;
 
@@ -1189,10 +1190,7 @@ int proxy(self_t *self) {
 
     /* test if main is back */
     
-    if ((status = worker_new(&client, "", "                        ", &global, 
-	                     NULL)) != APR_SUCCESS) {
-      return status;
-    } 
+    worker_new(&client, "", &global, NULL);
 
     worker_get_socket(client, "Default", "0");
     client->socket->socket_state = listener->socket->socket_state;
@@ -1262,6 +1260,9 @@ int main(int argc, const char *const argv[]) {
   const char *end_file = NULL;
   const char *cookie_pre = "COOKIE_";
   int flags = SELF_FLAGS_NONE;
+  apr_file_t *out;
+  apr_file_t *err;
+  appender_t *appender;
 
   srand(apr_time_now()); 
   
@@ -1273,6 +1274,9 @@ int main(int argc, const char *const argv[]) {
   apr_signal_block(SIGPIPE);
 #endif
   
+  apr_file_open_flags_stderr(&err, APR_BUFFERED|APR_XTHREAD, pool);
+  apr_file_open_flags_stdout(&out, APR_BUFFERED|APR_XTHREAD, pool);
+
   /* get options */
   apr_getopt_init(&opt, pool, argc, argv);
   while ((status = apr_getopt_long(opt, options, &c, &optarg)) 
@@ -1413,7 +1417,9 @@ int main(int argc, const char *const argv[]) {
   /* overwrites config */
   self->pool = pool;
   self->port = port;
-  self->log_mode = log_mode;
+  self->logger = logger_new(pool, log_mode, 0);
+  appender = appender_std_new(pool, out, APPENDER_STD_NONE);
+  logger_set_appender(self->logger, appender, "std", LOG_NONE, LOG_DEBUG);
   self->timeout = apr_pstrdup(pool, tmo);
   self->host_var = host_var ? apr_pstrdup(pool, host_var) : NULL;
   self->port_var = port_var ? apr_pstrdup(pool, port_var) : NULL;
