@@ -18,9 +18,12 @@
 extern module_t modules[];
 
 /* MODULES_DECLARATION */
+apr_status_t xml_module_init(global_t *global);
 apr_status_t html_module_init(global_t *global);
 apr_status_t js_module_init(global_t *global);
 apr_status_t lua_module_init(global_t *global);
+apr_status_t annotation_module_init(global_t *global);
+apr_status_t charset_module_init(global_t *global);
 apr_status_t perf_module_init(global_t *global);
 apr_status_t dbg_module_init(global_t *global);
 apr_status_t sys_module_init(global_t *global);
@@ -36,9 +39,12 @@ apr_status_t ssl_module_init(global_t *global);
 
 module_t modules[] = {
   /* MODULES_REGISTRATION */
+  { xml_module_init },
   { html_module_init },
   { js_module_init },
   { lua_module_init },
+  { annotation_module_init },
+  { charset_module_init },
   { perf_module_init },
   { dbg_module_init },
   { sys_module_init },
