@@ -26,6 +26,7 @@
 #define HTTEST_WORKER_H
 
 #include <apr_hooks.h>
+#include "logger.h" 
 #include "transport.h" 
 #include "store.h" 
 
@@ -42,9 +43,12 @@ typedef struct socket_s {
   char peek[32];
   apr_table_t *cookies;
   char *cookie;
+  sockreader_t *sockreader;
 } socket_t;
 
 typedef struct validation_s {
+  apr_table_t *ns;
+
   apr_table_t *dot;
   apr_table_t *headers;
   apr_table_t *body;
@@ -52,21 +56,6 @@ typedef struct validation_s {
   apr_table_t *exec;
 } validation_t;
 
-typedef struct recorder_s {
-  int on;
-#define RECORDER_OFF 0
-#define RECORDER_RECORD 1
-#define RECORDER_PLAY 2
-  int flags;
-#define RECORDER_RECORD_NONE 0
-#define RECORDER_RECORD_STATUS 1
-#define RECORDER_RECORD_HEADERS 2
-#define RECORDER_RECORD_BODY 4
-#define RECORDER_RECORD_ALL RECORDER_RECORD_STATUS|RECORDER_RECORD_HEADERS|RECORDER_RECORD_BODY 
-  apr_pool_t *pool;
-  sockreader_t *sockreader;
-} recorder_t;
-
 typedef struct worker_s worker_t;
 typedef struct global_s global_t;
 typedef apr_status_t(*interpret_f)(worker_t * self, worker_t *parent, 
@@ -93,6 +82,11 @@ struct worker_s {
   store_t *retvars;
   /* block local variables */
   store_t *locals;
+  /* buffered stdout */
+  apr_file_t *out;
+  /* buffered errout */
+  apr_file_t *err;
+  /* filename of current script part */
   const char *filename;
   apr_file_t *tmpf;
 #define FLAGS_NONE           0x00000000
@@ -116,8 +110,8 @@ struct worker_s {
   int cmd_from;
   int cmd_to;
   int which;
+  int group;
   char *name;
-  char *prefix;
   char *additional;
   char *file_and_line;
   const char *short_desc;
@@ -128,8 +122,8 @@ struct worker_s {
   char *match_seq;
   apr_time_t socktmo;
   apr_thread_t *mythread;
-  apr_thread_cond_t *sync_cond;
   apr_thread_mutex_t *sync_mutex;
+  apr_thread_mutex_t *log_mutex;
   apr_thread_mutex_t *mutex;
   apr_table_t *lines;
   apr_table_t *cache;
@@ -148,19 +142,7 @@ struct worker_s {
   socket_t *socket;
   apr_port_t listener_port;
   char *listener_addr;
-  sockreader_t *sockreader;
-  recorder_t *recorder;
-#define LOG_NONE 0
-#define LOG_ERR 1
-#define LOG_WARN 2
-#define LOG_INFO 3
-#define LOG_CMD 4
-#define LOG_ALL_CMD 5
-#define LOG_DEBUG 6
-  int log_mode;
-#if APR_HAS_FORK
-  apr_hash_t *procs;
-#endif
+  logger_t *logger;
 };
 
 struct global_s {
@@ -174,15 +156,17 @@ struct global_s {
   apr_hash_t *modules;
   apr_hash_t *blocks;
   apr_table_t *files;
-  int log_mode;
   apr_table_t *threads;
   apr_table_t *clients;
   apr_table_t *servers;
   apr_table_t *daemons;
-  int CLTs; 
-  int SRVs; 
-  apr_thread_cond_t *cond; 
-  apr_thread_mutex_t *sync;
+  logger_t *logger;
+  int CLTs;
+  int SRVs;
+  int cur_threads; 
+  int tot_threads; 
+  int groups;
+  apr_thread_mutex_t *sync_mutex;
   apr_thread_mutex_t *mutex;
   int line_nr;
 #define GLOBAL_STATE_NONE   0
@@ -196,7 +180,6 @@ struct global_s {
 #define GLOBAL_FILE_STATE_MODULE 1
   int file_state;
   int socktmo;
-  char *prefix;
   worker_t *worker;
   worker_t *cur_worker;
   apr_threadattr_t *tattr;
@@ -340,6 +323,8 @@ APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, block_end,
                           (global_t *global));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, client_create,
                           (worker_t *worker, apr_thread_start_t func, apr_thread_t **new_thread));
+APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, server_create,
+                          (worker_t *worker, apr_thread_start_t func, apr_thread_t **new_thread));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, worker_finally,
                           (worker_t *worker));
 APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, thread_start,
@@ -352,13 +337,6 @@ APR_DECLARE_EXTERNAL_HOOK(htt, HTT, apr_status_t, worker_joined,
 apr_status_t transport_register(socket_t *socket, transport_t *transport);
 apr_status_t transport_unregister(socket_t *socket, transport_t *transport);
 
-/** helpers */
-apr_status_t worker_new(worker_t ** self, char *additional,
-                        char *prefix, global_t *global, interpret_f interpret);
-apr_status_t worker_clone(worker_t ** self, worker_t * orig); 
-apr_status_t worker_handle_buf(worker_t *worker, apr_pool_t *pool, char *buf, 
-                               apr_size_t len); 
-
 /** commands */
 apr_status_t command_CALL(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp); 
 apr_status_t command_REQ(command_t * self, worker_t * worker, char *data, apr_pool_t *ptmp);
@@ -386,6 +364,7 @@ apr_status_t command_SOCKSTATE(command_t * self, worker_t * worker, char *data,
 apr_status_t command_HEADER(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp);
 apr_status_t command_RAND(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp);
 apr_status_t command_DEBUG(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp);
+apr_status_t command_PRINT(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp);
 apr_status_t command_UP(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp);
 apr_status_t command_DOWN(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp);
 apr_status_t command_TIME(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp);
@@ -411,9 +390,6 @@ apr_status_t command_BREAK(command_t *self, worker_t *worker, char *data, apr_po
 apr_status_t command_PRINT_HEX(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp); 
 apr_status_t command_AUTO_CLOSE(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp); 
 apr_status_t command_AUTO_COOKIE(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp); 
-#if APR_HAS_FORK
-apr_status_t command_PROC_WAIT(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp); 
-#endif
 apr_status_t command_MATCH_SEQ(command_t * self, worker_t * worker, char *data, apr_pool_t *ptmp);
 apr_status_t command_RECORD(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp); 
 apr_status_t command_PLAY(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp); 
@@ -426,10 +402,17 @@ apr_status_t command_VERSION(command_t *self, worker_t *worker, char *data, apr_
 apr_status_t command_DUMMY(command_t *self, worker_t *worker, char *data, apr_pool_t *ptmp); 
 
 /** helper */
-void worker_log(worker_t * self, int log_mode, char *fmt, ...); 
-void worker_log_error(worker_t * self, char *fmt, ...); 
-void worker_log_buf(worker_t * self, int log_mode, const char *buf, char *prefix, 
-                    int len); 
+void lock(apr_thread_mutex_t *mutex); 
+void unlock(apr_thread_mutex_t *mutex); 
+void worker_new(worker_t ** self, char *additional,
+                global_t *global, interpret_f interpret);
+void worker_clone(worker_t ** self, worker_t * orig); 
+apr_status_t worker_handle_buf(worker_t *worker, apr_pool_t *pool, char *buf, 
+                               apr_size_t len); 
+
+void worker_log(worker_t * worker, int mode, char *fmt, ...); 
+void worker_log_buf(worker_t * worker, int mode, char dir, const char *buf,
+                    apr_size_t len); 
 void worker_var_set(worker_t * worker, const char *var, const char *val); 
 const char * worker_var_get(worker_t * worker, const char *var); 
 void worker_test_reset(worker_t * worker); 
@@ -443,7 +426,6 @@ const char * worker_resolve_var(worker_t *worker, const char *name, apr_pool_t *
 char * worker_replace_vars(worker_t * worker, char *line, int *unresolved,
                            apr_pool_t *ptmp); 
 apr_status_t worker_flush(worker_t * self, apr_pool_t *ptmp);
-apr_status_t worker_clone(worker_t ** self, worker_t * orig); 
 void worker_destroy(worker_t * self); 
 apr_status_t worker_match(worker_t * worker, apr_table_t * regexs, 
                           const char *data, apr_size_t len); 
